
/*
 * This modification of the original file JDBCSampler.java  
 * taken from the Apache SVN repository. The starting point from this file
 * was rev 412837 of JDBCSamplerBeanInfo.java  
 */


/*
 * Copyright 2001-2006 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 */

package com.rubenlaguna.jmeter.jdbc.sampler;

import java.lang.reflect.Field;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;

import org.apache.avalon.excalibur.datasource.DataSourceComponent;
import org.apache.jmeter.samplers.AbstractSampler;
import org.apache.jmeter.samplers.Entry;
import org.apache.jmeter.samplers.SampleResult;
import org.apache.jmeter.testbeans.TestBean;
import org.apache.jorphan.collections.Data;
import org.apache.jorphan.logging.LoggingManager;
import org.apache.log.Logger;

/**
 * A sampler which understands JDBC database requests.
 * 
 * @author Original author unknown
 * @author <a href="mailto:jeremy_a@bigfoot.com">Jeremy Arnold</a>
 */
public class EnhancedJDBCSampler extends AbstractSampler implements TestBean {
	private static final int MAX_ENTRIES = 200;

	private static final Logger log = LoggingManager.getLoggerForClass();
    static Map mapJdbcNameToInt;

    // Query types (used to communicate with GUI)
	static final String SELECT   = "Select Statement";
	static final String UPDATE   = "Update Statement";
	static final String CALLABLE = "Callable Statement";
	static final String PREPARED_SELECT = "Prepared Select Statement";
	static final String PREPARED_UPDATE = "Prepared Update Statement";
	static final String COMMIT = "Commit";

	private String query = "";

	private String dataSource = "";

	private String queryType = SELECT;
	private String queryArguments = "";
	private String queryArgumentsTypes = "";
	
	/**
	 *  Cache of PreparedStatements stored in a per-connection basis. Each entry of this 
	 *  cache is another Map mapping the statemente string to the actual PreparedStatement.
	 *  The cache has a fixed size of MAX_ENTRIES and it will throw aways all PreparedStatements 
	 *  from the least recently used connections.  
	 */
	private static Map perConnCache = new LinkedHashMap(MAX_ENTRIES){
		protected boolean removeEldestEntry(java.util.Map.Entry arg0) {
			if (size() > MAX_ENTRIES) {
				final Object value = arg0.getValue();
				if (value instanceof Map) {
					closeAllStatements(((Map)value).values());
				}
				return true;
			}
			return false;
		}
	};
	
//	private static Map connPerThreadCache = new HashMap();

	/**
	 * Creates a JDBCSampler.
	 */
	public EnhancedJDBCSampler() {
	}

	public SampleResult sample(Entry e) {
		log.debug("sampling jdbc");
		
        SampleResult res = new SampleResult();
		res.setSampleLabel(getName());
		res.setSamplerData(toString());
        res.setDataType(SampleResult.TEXT);
        // Bug 31184 - make sure encoding is specified
        res.setDataEncoding(System.getProperty("file.encoding"));

        // Assume we will be successful
        res.setSuccessful(true);


		res.sampleStart();
		DataSourceComponent pool = (DataSourceComponent) getThreadContext().getVariables().getObject(getDataSource());
		log.debug("DataSourceComponent: " + pool);
		Connection conn = null;
		Statement stmt = null;
		

		try {

			if (pool == null)
				throw new SQLException("No pool created");

			
			if (conn == null) {
				conn = pool.getConnection();
			}
			log.debug("connection id " + conn + " autocommit enabled: " +conn.getAutoCommit());

            // Based on query return value, get results
            String _queryType = getQueryType();
            if (SELECT.equals(_queryType)) {
            	stmt = conn.createStatement();
				ResultSet rs = null;
				try {
					rs = stmt.executeQuery(getQuery());
					Data data = getDataFromResultSet(rs);
					res.setResponseData(data.toString().getBytes());
				} finally {
					close(rs);
				}
            } else if (CALLABLE.equals(_queryType)) {
            	CallableStatement cstmt = getCallableStatement(conn);
            	setArguments(cstmt);
            	// A CallableStatement can return more than 1 ResultSets
            	// plus a number of update counts. 
            	boolean hasResultSet = cstmt.execute();
            	String sb = resultSetsToString(cstmt,hasResultSet);
            	res.setResponseData(sb.toString().getBytes());       	
            } else if (UPDATE.equals(_queryType)) {
            	stmt = conn.createStatement();
            	stmt.executeUpdate(getQuery());
				int updateCount = stmt.getUpdateCount();
				String results = updateCount + " updates";
				res.setResponseData(results.getBytes());
            } else if (PREPARED_SELECT.equals(_queryType)) {
            	PreparedStatement pstmt = getPreparedStatement(conn);
            	setArguments(pstmt);
            	pstmt.executeQuery();
            	String sb = resultSetsToString(pstmt,true);
            	res.setResponseData(sb.toString().getBytes());
            } else if (PREPARED_UPDATE.equals(_queryType)) {
            	PreparedStatement pstmt = getPreparedStatement(conn);
            	setArguments(pstmt);
            	pstmt.executeUpdate();
				String sb = resultSetsToString(pstmt,false);
            	res.setResponseData(sb.toString().getBytes());
            } else if (COMMIT.equals(_queryType)) {
            	res.setResponseData("committed".getBytes());
            	perConnCache.remove(getThreadName());            	
            } else { // User provided incorrect query type
                String results="Unexpected query type: "+_queryType;
                res.setResponseMessage(results);
                res.setSuccessful(false);
			}

		} catch (SQLException ex) {
			log.error("Error in JDBC sampling", ex);
			if (ex.getMessage().contains("is closed")) {
				try {
					conn.close();
					perConnCache.remove(getThreadName());
				} catch (SQLException e1) {				
					log.error("Error in JDBC sampling", e1);
				}
			}
			res.setResponseMessage(ex.toString());
			res.setSuccessful(false);
		} finally {
			close(stmt);
			close(conn);
		}

		res.sampleEnd();
		return res;
	}

	private String resultSetsToString(PreparedStatement pstmt, boolean result) throws SQLException {
		StringBuffer sb = new StringBuffer();
		sb.append("\n");
		int updateCount = 0;
		if (!result) {
			updateCount = pstmt.getUpdateCount();
		}
		do {
			if (result) {
				ResultSet rs = null;
				try {
					rs = pstmt.getResultSet();
					Data data = getDataFromResultSet(rs);
					sb.append(data.toString()).append("\n");
				} finally {
					close(rs);
				}
			} else {
				sb.append(updateCount).append(" updates.\n");
			}
			result = pstmt.getMoreResults();
			if (!result) {
				updateCount = pstmt.getUpdateCount();
			}
		} while (result || (updateCount != -1));
		return sb.toString();
	}


	private void setArguments(PreparedStatement pstmt) throws SQLException {
		if ("".equals(getQueryArguments().trim())) {
			return;
		}
		String[] arguments = getQueryArguments().split(",");
		String[] argumentsTypes = getQueryArgumentsTypes().split(",");
		if (arguments.length != argumentsTypes.length) {
			throw new SQLException("number of arguments ("+arguments.length+") and number of types ("+argumentsTypes.length+") are not equal");
		}
		for (int i = 0; i < arguments.length; i++) {
			String argument = arguments[i];
			String argumentType = argumentsTypes[i];
			//TODO should be a more elegant way to do it
		    int targetSqlType = getJdbcType(argumentType);
		    pstmt.setObject(i+1, argument, targetSqlType);
		}
	}
    
    public static int getJdbcType(String jdbcType) throws SQLException {
        // based on javaalmanac e291. Getting the Name of a JDBC Type from javaalmanac.com
        // http://javaalmanac.com/egs/java.sql/JdbcInt2Str.html
    	if (mapJdbcNameToInt == null) {
    		mapJdbcNameToInt = new HashMap();
    		
    		//Get all field in java.sql.Types
    		Field[] fields = java.sql.Types.class.getFields();
            for (int i=0; i<fields.length; i++) {
                try {
                    // Get field name
                	log.info("ruben");
                	if (fields[i] != null) {
                		String name = fields[i].getName();

                		// Get field value
                		Integer value = (Integer)fields[i].get(null);

                		// Add to map
                		mapJdbcNameToInt.put(name.toLowerCase(),value);
                	}
                } catch (IllegalAccessException e) {
                	throw new RuntimeException(e);
                }
            }
            mapJdbcNameToInt.put("string", new Integer(java.sql.Types.VARCHAR));
            mapJdbcNameToInt.put("str", new Integer(java.sql.Types.VARCHAR));
            mapJdbcNameToInt.put("int", new Integer(java.sql.Types.INTEGER));
            mapJdbcNameToInt.put("integer", new Integer(java.sql.Types.INTEGER));
    		
    	}
    	final Object intValue = mapJdbcNameToInt.get(jdbcType.toLowerCase().trim());
    	if (null == intValue) {
    		throw new SQLException("type " + jdbcType + " cannot be mapped to any JDBC data type");
    	}
		return ((Integer)intValue).intValue();
    }
	

	private CallableStatement getCallableStatement(Connection conn) throws SQLException {
		return (CallableStatement) getPreparedStatement(conn,true);
		
	}
	private PreparedStatement getPreparedStatement(Connection conn) throws SQLException {
		return getPreparedStatement(conn,false);
	}

	private PreparedStatement getPreparedStatement(Connection conn, boolean callable) throws SQLException {
		Map preparedStatementMap = (Map) perConnCache.get(conn);
		if (null == preparedStatementMap ) {
		    // MRU PreparedStatements cache. 
			preparedStatementMap = new LinkedHashMap(MAX_ENTRIES) {
				protected boolean removeEldestEntry(java.util.Map.Entry arg0) {
					final int theSize = size();
					if (theSize > MAX_ENTRIES) {
						Object value = arg0.getValue();
						if (value instanceof PreparedStatement) {
							PreparedStatement pstmt = (PreparedStatement) value;
							try {
								pstmt.close();
							} catch (SQLException e) {
								// ignore this exception
								e.printStackTrace();
							}
						}
						return true;
					}
					return false;
				}
			};
			perConnCache.put(conn, preparedStatementMap);
		}
		PreparedStatement pstmt = (PreparedStatement) preparedStatementMap.get(getQuery());
		if (null == pstmt) {
			if (callable) {
				pstmt = conn.prepareCall(getQuery());
			} else {
				pstmt = conn.prepareStatement(getQuery());
			}
			preparedStatementMap.put(getQuery(), pstmt);
		}
		pstmt.clearParameters();
		return pstmt;
	}

	private static void closeAllStatements(Collection collection) {
		Iterator iterator = collection.iterator();
		while (iterator.hasNext()) {
			PreparedStatement pstmt = (PreparedStatement) iterator.next();
			try {
				pstmt.close();
			} catch (SQLException e) {
				// ignore this exception
				e.printStackTrace();
			}
		}
		
	}

	/**
	 * Gets a Data object from a ResultSet.
	 * 
	 * @param rs
	 *            ResultSet passed in from a database query
	 * @return a Data object
	 * @throws java.sql.SQLException
	 */
	private Data getDataFromResultSet(ResultSet rs) throws SQLException {
		ResultSetMetaData meta = rs.getMetaData();
		Data data = new Data();

		int numColumns = meta.getColumnCount();
		String[] dbCols = new String[numColumns];
		for (int i = 0; i < numColumns; i++) {
			dbCols[i] = meta.getColumnName(i + 1);
			data.addHeader(dbCols[i]);
		}

		while (rs.next()) {
			data.next();
			for (int i = 0; i < numColumns; i++) {
				Object o = rs.getObject(i + 1);
				if (o instanceof byte[]) {
					o = new String((byte[]) o);
				}
				data.addColumnValue(dbCols[i], o);
			}
		}
		return data;
	}
	
	/* returns the connection to the pool. Connection is not really closed */
	public static void close(Connection c) {
		try {
			if (c != null) {
				if (c.getAutoCommit() == false) {
					c.commit(); /* if we're not in autocommit mode then commit now */ 
				}
				c.close();
			}
		} catch (SQLException e) {
			log.warn("Error closing Connection", e);
		}
	}
	
	public static void close(Statement s) {
		try {
			if (s != null) s.close();
		} catch (SQLException e) {
			log.warn("Error closing Statement " + s.toString(), e);
		}
	}

	
	public static void close(ResultSet rs) {
		try {
			if (rs != null) rs.close();
		} catch (SQLException e) {
			log.warn("Error closing ResultSet", e);
		}
	}

	public String getQuery() {
		return query;
	}

	public String toString() {
        StringBuffer sb = new StringBuffer(80);
        sb.append("[");
        sb.append(getQueryType());
        sb.append("] ");
        sb.append(getQuery());
		return sb.toString();
	}

	/**
	 * @param query
	 *            The query to set.
	 */
	public void setQuery(String query) {
		this.query = query;
	}

	/**
	 * @return Returns the dataSource.
	 */
	public String getDataSource() {
		return dataSource;
	}

	/**
	 * @param dataSource
	 *            The dataSource to set.
	 */
	public void setDataSource(String dataSource) {
		this.dataSource = dataSource;
	}

	/**
	 * @return Returns the queryType.
	 */
	public String getQueryType() {
		return queryType;
	}

	/**
	 * @param queryType The queryType to set.
	 */
	public void setQueryType(String queryType) {
		this.queryType = queryType;
	}

	public String getQueryArguments() {
		return queryArguments;
	}

	public void setQueryArguments(String queryArguments) {
		this.queryArguments = queryArguments;
	}

	public String getQueryArgumentsTypes() {
		return queryArgumentsTypes;
	}

	public void setQueryArgumentsTypes(String queryArgumentsType) {
		this.queryArgumentsTypes = queryArgumentsType;
	}
}
